! this software is distributed under the GNU license version 2
module kinds
  integer, parameter :: single = SELECTED_REAL_KIND( 6, 37 )
  integer, parameter :: double = SELECTED_REAL_KIND( 15, 307 )
  integer, parameter :: extended = SELECTED_REAL_KIND( 18, 4931 )
  integer, parameter :: r4 = SELECTED_REAL_KIND( 6, 37 )
  integer, parameter :: r8 = SELECTED_REAL_KIND( 15, 307 )
  integer, parameter :: r16 = SELECTED_REAL_KIND( 18, 4931 )

  ! current precison for hash storage
  integer, parameter :: rh=r8
end module kinds


module util2
use kinds
real(r8)::denseop_tol=1.d-10
contains


SUBROUTINE PRINTM8 (matrix,text,fmt,un)
! from IM blupf90, modified to print matrix of any shape
! prints matrix and text(optionaly) using form (optionally) to standard
! output or a unit(optionally)
character(*), optional::text*(*),fmt
character(40)::fmt1
integer :: n,i
integer,optional::un
real (r8):: matrix(:,:)
if (present(text)) then
   if (present(un)) then
         write(un,'(a)')text
      else
         print*,text; 
   endif
end if
if (present(fmt)) then; fmt1=fmt;else
fmt1='(10g12.4/)';end if
n=(size(matrix,1))
do i=1,n
   if (present(un)) then
         write(un,fmt1)matrix(i,:)
      else
         print fmt1,matrix(i,:)
   endif
enddo
END SUBROUTINE PRINTM8

  function number_of_lines(file) result(n)
  implicit none
  integer::file,n,io
  ! the file is supposed to exist and be already opened
  n=0
  rewind(file)
  do
        read(file,*,iostat=io)
        if(io/=0) exit
        n=n+1
  enddo
  rewind(file)
end function

function diag_mat_vec8(x) result (y)
! y=diag(x) if x is a matrix and y a vector
real (r8)::x(:,:)
real (r8)::y(size(x,1))
integer::i
!
do i=1,size(y)
   y(i)=x(i,i)
enddo
end function   


SUBROUTINE CHOLS8(x,rank)
! cholesky decomposition
integer :: n,i,j,k,rank_o
integer, intent(out), optional::rank
real(r8)::diagsq,d_p
real(r8)::x(:,:)
rank_o=0
n=size(x,1)
do i=1,n
   d_p=0
   do k=1,i-1
      d_p=d_p+x(i,k)*x(i,k)
   enddo
!   diagsq=x(i,i)-dot_product(x(i,1:i-1),x(i,1:i-1))
   diagsq=x(i,i)-d_p
  if (abs(diagsq).lt.denseop_tol) then
        x(i,:)=0;x(:,i)=0       !zero row and column
      elseif (diagsq.lt.0) then
        print*,' Matrix not semipositive-definite, row ',i
        stop
!   endif
   else
    rank_o=rank_o+1
   x(i,i)=sqrt(diagsq)
   do j=i+1,n     
      d_p=0
      do k=1,i-1
         d_p=d_p+x(j,k)*x(i,k)
      end do
!      x(j,i)=(x(j,i)-dot_product(x(j,1:i-1),x(i,1:i-1)))/x(i,i)
      x(j,i)=(x(j,i)-d_p)/x(i,i)
      x(i,j)=x(j,i)
   enddo
   end if
enddo
! zero upper-diagonals

do i=1,n
   x(i,i+1:n)=0
enddo   

if (present(rank))then
 rank=rank_o
end if

END SUBROUTINE CHOLS8


FUNCTION FCHOLS8(X)  RESULT(Y)
 real(r8)::X(:,:),y(size(x,1),size(x,1))
 y=x
 call chols8(y)
END FUNCTION FCHOLS8

SUBROUTINE INVCS8(l,rank)
! calculates inverse of LL', where L is cholesky decomposition 
!USE LA_PRECISION, ONLY: WP => DP
integer  ::n,i,j,k,rank_o
integer, intent(out), optional::rank
real(r8) :: l(:,:),d_p
real(r8) ::w(size(l,1))
n=size(l,1)

rank_o=0

call chols8(l,rank_o)   ! cholesky factorization

do i=1,n
   w(i:n)=0
   if (abs(l(i,i)).gt.denseop_tol) w(i)=1/l(i,i)
   ! forward substitution
   do j=i+1,n
      d_p=0
      do k=i,j-1
         d_p=d_p+l(j,k)*w(k)
      end do
!      if (abs(l(j,j)).gt.denseop_tol)
!      w(j)=-dot_product(l(j,i:j-1),w(i:j-1))/l(j,j)
      if (abs(l(j,j)).gt.denseop_tol) w(j)=-d_p/l(j,j)
   enddo
   !backward substitution
   do j=n,i,-1
      d_p=0
      do k=j+1,n
         d_p=d_p+l(k,j)*w(k)
      end do
!      if (abs(l(j,j)).gt.denseop_tol)
!      w(j)=(w(j)-dot_product(l(j+1:n,j),w(j+1:n)))/l(j,j)
      if (abs(l(j,j)).gt.denseop_tol) w(j)=(w(j)-d_p)/l(j,j)
   enddo
   l(i:n,i)=w(i:n)
   l(i,i:n)=w(i:n)
enddo   

if (present(rank)) then
 rank=rank_o
end if

END SUBROUTINE INVCS8
FUNCTION FINVCS8(X) RESULT(Y)
 real(r8)::X(:,:),y(size(x,1),size(x,1))
 y=x
 call invcs8(y)
END FUNCTION FINVCS8



end module util2


module pedigree
use util2
use kinds
integer,allocatable:: ped(:,:)
real(r8),allocatable:: gamma(:,:),A(:,:),Ainv(:,:),F(:),D(:),qgqp(:,:)
integer:: nanim,nmf ! number of metafounders

contains

function seconds()
     call cpu_time(seconds)
end function

subroutine read_ped(file)
        implicit none
        integer:: file,i

        nanim= number_of_lines(file)
        allocate(ped(nanim,3))
        do i =1,nanim
                read(file,*) ped(i,:)
        enddo
end subroutine

subroutine read_gamma(file)
        ! assumes read_ped has been called
        implicit none
        integer:: i,file

        nmf=count(ped(:,2)==0)
        allocate(gamma(nmf,nmf))
        do i=1,nmf
                read(file,*) gamma(i,:)
        enddo
end subroutine        



subroutine emik_and_terrill()
        ! assumes read_gamma has been called
        implicit none
        integer:: i,j

        if(.not.allocated(A))allocate(A(nanim,nanim))
        A=0
        A(1:nmf,1:nmf)=gamma
        do i=1,nanim
                if(i>nmf ) A(i,i)=1+0.5*A(ped(i,2),ped(i,3))
                do j=1,i-1
                        if(i>nmf .or. j>nmf) then
                                ! not the metafounder square, use rules
                                A(i,j)=0.5*(A(ped(i,2),j)+A(ped(i,3),j))
                                A(j,i)=A(i,j)
                        endif
                enddo
        enddo
end subroutine

subroutine allocF()
if(.not.allocated(F)) then
        allocate(F(nanim))
endif
F=-5 ! not computed (this cannot be -1 because it is -1 if unrelated)
end subroutine

subroutine allocD()
if(.not.allocated(D)) then
        allocate(D(nanim))
endif
D=0
end subroutine

subroutine allocAinv()
if(.not.allocated(Ainv)) then
        allocate(Ainv(nanim,nanim))
endif
Ainv=0
end subroutine
subroutine allocQgqp()
if(.not.allocated(qgqp)) then
        allocate(qgqp(nanim,nanim))
endif
qgqp=0
end subroutine

subroutine getFfromA()
        implicit none
        ! assumes A computed
        if(.not.allocated(A)) then
                print *, 'A not allocated, stop'
                stop
        endif
        call allocF()
        F=diag_mat_vec8(A)-1
end subroutine

subroutine getDfromF()
        ! this computes D *only* for real animals (not metafounders) in which case D is diagonal
        implicit none
        integer:: i
        ! assumes F computed
        if(.not.allocated(F).or. any(F==-5)) then
                print *, 'F not allocated or computed, stop'
                stop
        endif
        call allocD()
        do i=nmf+1,nanim
                D(i)=0.5-0.25*(F(ped(i,2))+F(ped(i,3)))
        enddo
end subroutine

recursive function cffa(a1,a2) result(s)
! calculates recursively coefficients for pair of animals
! from I Aguilar inbupgf90
   implicit none
   integer:: a1,a2
   real(r8):: s
   if (a1<=nmf .and. a2<=nmf) then
        ! I dont' know if this works when an individual is half known
      s= gamma(a1,a2)
   elseif (a1==a2) then
      if (f(a1)==-5) then
        ! if F not computed for this individual
         f(a1) = 0.5*cffa(ped(a1,2),ped(a1,3))
      endif
      s = 1 + f(a1)
   !elseif (a1<=nmf .and. a2>nmf) then
   !elseif (a1<=nmf .and. a2>nmf) then

   else
      if (a1<a2)  then 
         s=0.5*(cffa(a1,ped(a2,2)) + cffa(a1,ped(a2,3)))
      else
         s=0.5*(cffa(a2,ped(a1,2)) + cffa(a2,ped(a1,3)))
      endif
   endif
end function

subroutine getFrecursive()
        implicit none
        integer:: i
        call allocF()
        do i=1,nmf
                F(i)=gamma(i,i)-1
        enddo
        do i=nmf+1,nanim
                F(i)=0.5*cffa(ped(i,2),ped(i,3))
        enddo
end subroutine





subroutine meuw_related() 
   !  Meuwissen & Luo (1992) for related metafounders, modifications by AL
   implicit none
   integer :: ss(nanim),dd(nanim)
   integer, allocatable:: point(:)
   real(r8), allocatable:: l(:),d(:)
   integer :: is,id,i,j,k,n,ks,kd
   real(r8) :: r,fi
   real :: t1
   
   real(r8):: Lgamma(nmf,nmf) ! gamma = Lgamma Lgamma' (gamma=Abas in OFC notation)
   logical:: not_ml=.true.


   !print*,'Calculating Inbreeding by M&L function'
   t1=seconds()
   call allocF()
   Lgamma=fchols8(gamma) !lower triangular L in LL' 
   !call printm8(Lgamma,text='Lgamma')

   ss=ped(:,2); dd=ped(:,3)
   n=size(ss,dim=1)
   allocate(point(n),l(n),d(n))
   point=0;l=0;d=0
   do i=1,nmf
        f(i)=gamma(i,i)-1
   enddo
   do i=nmf+1,n
      if(mod(i,int(n/1))==0) &
        write(*,*) ' at',i,' animals'
      is=ss(i); id=dd(i)
      ss(i)=max(is,id)
      dd(i)=min(is,id)
      d(i)=.5-.25*(f(is)+f(id))
      if (is==0 .or. id==0) then
         write (*,*) 'missing genealogy for non-metafounder',i,is,is
         stop
         cycle
      endif
      fi=-1 !! this is the -1 in inbreeding=Aii-1
      l(i)=1.0
      j=i
      do while (j/=0)
         k=j
         r=0.5*l(k)
         ks=ss(k)
         kd=dd(k)
         if (ks>0) then
            do while(point(k)>ks)
               k=point(k)
            enddo
            l(ks)=l(ks)+r
            if (ks/=point(k))then
               point(ks)=point(k)
               point(k)=ks
            endif
            if (kd>0)then
               do while (point(k)>kd)
                  k=point(k)
               enddo
               l(kd)=l(kd)+r
               if (kd/=point(k)) then
                  point(kd)=point(k)
                  point(k)=kd
               endif
            endif
         endif
         ! need to accomodate here the fact if that one animal may be born from TWO metafounders
         ! in which case it is not easy how to process sires and dams separately

         ! sum l(j,t) for all combinations (j,t) in metafounders
         ! hence we need to keep some j-terms in L (those that are metafounders)
         if(not_ml) then
           if(j>nmf) then
              fi=fi+l(j)*l(j)*d(j)
              l(j)=0 ! otherwise I keep them in memory
           endif
         else
           ! old M&L code
           fi=fi+l(j)*l(j)*d(j)
           l(j)=0
         endif
         k=j
         j=point(j)
         point(k)=0
      enddo
      ! now add contributions from metafounders
      if(not_ml) then
        fi=fi+sum(matmul(l(1:nmf),Lgamma)**2)
        l(1:nmf)=0
      endif
      f(i)=fi
   enddo
   print*,'   Calculating Inbreeding by M&L, elapsed time',seconds()-t1
end subroutine



subroutine getAgammaInverse(scaled)
        implicit none
        real(r8),allocatable:: Abasinv(:,:)
        real(r8):: k
        integer:: i
        logical,optional:: scaled
        !this subroutine computes the Agamma-inverse following two parts:
        ! 1-inversion of the metafounders ( give 0 if unrelated)
        ! 2-contribution from offspring
        ! if scaled=.true. then it is scaled in the variance of "unrelated"
        k=1d0
        if(present(scaled)) then
                !k=1+mean(diag(Gamma))/2-mean(Gamma)
                if(scaled) k=1+(sum(diag_mat_vec8(gamma))/nmf)/2 - sum(gamma)/(nmf**2)
                print *,'k correction factor, varu_related = varu_unrelated/k',k
        endif
        allocate(Abasinv(nmf,nmf))
        ! metafounders: can use generalized inverse from denseop (will consider correctly Gamma-matrix=0)
        Abasinv=finvcs8(gamma)
        call allocAinv()
        Ainv(1:nmf,1:nmf)=Abasinv*k
        ! otherwise, loop Abasinv to make controbutions using addm()
        ! offspring
        call add_g_add(k)
        deallocate(Abasinv)
end subroutine


  subroutine add_g_add(sc)
  ! IMisztal subroutine heavily modified to construct A-1 (without genetic variance!!)
! generates contributions for additive animal effect
  integer :: type,k,l,m,n,p(3)
  real(r8) ::w(3),val,mendel,sc
!
  w=(/1., -.5, -.5/)
  do i=nmf+1,nanim
     p=ped(i,:)
     ! notes from M A Elzo
     mendel=1d0/ d(i) 
     !if(ntotal<12)print *,'dii_ in:',i,d(i)
     do k=1,3
        do l=1,3        
           if (p(k) /= 0 .and. p(l) /=0) then
             m=p(k)
             n=p(l)
             !val=w(k)*w(l)*res(par_stat) version without inbreeding
             val=w(k)*w(l)*mendel
             !call addm(val,m,n,Ainv_hash) 
             Ainv(m,n)=Ainv(m,n)+val*sc
             !print *,i,m,n,val
           endif
        enddo
     enddo
  enddo
  end subroutine


subroutine A_times_v(w,v)
! Colleau 2002
! computes w=A*v using  TDT'v
! assumes ped and D(1:nanim) vector of inbreeding available
implicit none
integer :: i
real(r8) :: w(:),v(:),tmp,di
real(r8) ::q(size(v))
q=0

! solve first T
do i=nanim,1,-1
   q(i)=q(i)+v(i)
   if (ped(i,2)>0) q(ped(i,2)) = q(ped(i,2))+q(i)*0.5
   if (ped(i,3)>0) q(ped(i,3)) = q(ped(i,3))+q(i)*0.5
   !q(ped(i,:)) = q(ped(i,:))+q(i)*0.5
enddo
! multiply by D
! metafounders
q(1:nmf)=matmul(gamma,q(1:nmf))
! animals
q(nmf+1:nanim)=D(nmf+1:nanim)*q(nmf+1:nanim)
!solve again 
do i=1,nanim
   tmp = 0
   if (ped(i,2)>0) tmp = tmp +w(ped(i,2))
   if (ped(i,3)>0) tmp = tmp +w(ped(i,3))
   w(i)=.5*tmp
   w(i)= w(i) + q(i)
enddo
end subroutine

function getA22(list) result(A22)
! compute A22 for a list of individuals
implicit none
integer:: list(:)
real(r8):: A22(size(list),size(list)),x(size(list)),w(size(list))
integer:: n,i
	n=size(list)
	A22=0
	do i=1,n
		x=0; x(list(i))=1
		call A_times_v(w,x)
		A22(:,i)=w
	enddo
end function

subroutine Qp_times_x(v,x)
!! Computes v=Q'*x using Qb'Pb'(I-P)-t x ! ZGV version
! Computes v=Q'*x using (I-P)-t x
! from ZGV, modified by AL
! first ng individuals are metafounders
! the purpose of this subroutine is to get the sume of squares Q'ZZ'Q, etc; can be used to compute upg covariates
implicit none
integer :: i,is,id
real(8) :: x(:),v(:)
real(8) :: q(size(x))

q=0

! Computes q = (I-P)-t x
 ! note that, opposite to Zulma stuff, here UPGs are already included in I-P.
do i=nanim,1,-1
   q(i) = q(i) + x(i)
   is = ped(i,2); id = ped(i,3)
   if (is > 0) q(is) = q(is) + q(i)*0.5
   if (id > 0) q(id) = q(id) + q(i)*0.5
enddo
! on exit, the first ng are the contributions from animals to UPG
v=q

end subroutine

function getq(i) result(qrow)
implicit none
! compute nmf covariables associated to this individual (qrow=% of origin from
! metafounders)
integer:: i
real(r8):: qrow(nmf),x(nanim),v(nanim)
qrow=0
x=0; x(i)=1
call Qp_times_x(v,x)
qrow=v(1:nmf)
end function

subroutine qgammaqp() 
! computes QgammaQ' 
implicit none
integer:: i
real(r8):: q(nanim,nmf),v(nanim),x(nanim),qgq(nanim,nanim)
call allocQgqp()
q=0
do i=nmf+1,nanim
  !x=0; x(i)=1
  !call Qp_times_x(v,x)
  !q(i,:)=v(1:nmf)
  q(i,:)=getq(i)
enddo
qgqp=matmul(q,matmul(gamma,transpose(q)))
!call printm8(q)
end subroutine


end module pedigree

program sampleagamma
! example of Agamma related subroutines to construct the relationship matrix, 
! compute products Ax, compute inbreeding, and invert relationship matrix, all with metafounders
! code put together by Andres Legarra, INRA, 9 Oct 2014, based mostly on code by I
! Aguilar, I Misztal, ZG Vitezica, and himself 

use kinds
use util2
use pedigree
implicit none

character(len=100):: help='USAGE: agamma pedfile gammafile'
character(len=100):: pedfile,gammafile
character(len=10)::fmat='(200f6.2)' ! alternatively 200f7.4
integer:: io,i,j
real(r8),allocatable:: x(:),v(:)

if(command_argument_count()<2) then
        print *,help
        stop
endif        
call get_command_argument(1,value=pedfile,status=io)
call get_command_argument(2,value=gammafile,status=io)

open(unit=1,file=pedfile,status='old')
open(unit=2,file=gammafile,status='old')


call read_ped(1)
call read_gamma(2)

call emik_and_terrill()
write (*,*)'A^gamma from Emik & Terrill'
call printm8(A,fmt=fmat)
write (*,*)
call getFfromA()
write (*,*)
write (*,*)'F from Emik & Terrill'
write (*,fmat)F
call getDfromF()
write (*,*)
write (*,*)'D from F'
write (*,fmat)D
write (*,*)
!A=getA22((/1:nanim/)) Fortran 2008 I believe? works in ifort
A=getA22((/(i,i=1,nanim)/))
write (*,*)'A^gamma from Colleau'
call printm8(A,fmt=fmat)




call getFrecursive()
write (*,*)
write (*,*)'F from recursion'
write (*,fmat)F
call meuw_related()
write (*,*)'F from Meuwissen and Luo'
write (*,*)
write (*,fmat)F
call getDfromF()
write (*,*)
write (*,*)'D from F'
write (*,fmat)D
write (*,*)
call getAgammaInverse()
write (*,*)'A^gamma-1 using Henderson''s rules'
call printm8(Ainv,fmt=fmat)
write (*,*)
write (*,*)'A^gamma times A^gamma-1 = I'
write (*,*)'(but not with Gamma=0 as A^gamma is undefined for metafounders)'
call printm8(matmul(A,Ainv),fmt=fmat)

! the next lines show how to compute A-1 efficiently without setting up A
write (*,*)
write (*,*)'setting up A^gamma-1 *scaled* without setting up A^gamma'
deallocate(A,D,F,Ainv)
call meuw_related() ! call getFrecursive() is also efficient
call getDfromF()
call getAgammaInverse()
call getAgammaInverse(scaled=.TRUE.)
call printm8(Ainv,fmt=fmat)
write (*,*)

 

end









